import pytest
from mock import AsyncMock, MagicMock


def test_general_arguments():
    from frankfurt import fields

    with pytest.raises(TypeError):
        fields.UUIDField(pk=True)


@pytest.mark.asyncio
async def test_fields_by_examples():

    conn = AsyncMock()
    conn.transaction = MagicMock()

    from frankfurt import Database
    from frankfurt import fields

    db = Database()

    class Example_1(db.Model):
        password = fields.BinaryField()

    await db.create_all_tables(conn=conn)

    conn.execute.assert_awaited_once_with(
        'CREATE TABLE IF NOT EXISTS example_1 (password BYTEA)'
    )


def test_default_values():

    from frankfurt.models import Model
    from frankfurt import fields

    class Example(Model):
        text_1 = fields.CharField(max_length=200, default='text')
        text_2 = fields.CharField(max_length=200, default=None)
        text_3 = fields.CharField(max_length=200, default=lambda : "sentinel")

    example = Example()
    assert example['text_1'] == 'text'
    assert example['text_2'] is None
    assert example['text_3'] == 'sentinel'


def test_not_null_values():

    from frankfurt.models import Model
    from frankfurt import fields

    class Example(Model):
        text = fields.CharField(max_length=200, not_null=True)

    example = Example()

    with pytest.raises(KeyError):
        assert example['text'] is None


@pytest.mark.asyncio
async def test_unique_constraint():

    conn = AsyncMock()
    conn.transaction = MagicMock()

    from frankfurt import Database
    from frankfurt import fields

    db = Database()

    class Example(db.Model):
        token = fields.CharField(max_length=64, unique=True)

    await db.create_all_tables(conn=conn)

    conn.execute.assert_awaited_once_with(
        'CREATE TABLE IF NOT EXISTS example (token VARCHAR(64) UNIQUE)'
    )



from typing import Tuple, Type, Dict, Any, List

from frankfurt.fields import BaseField, ForeignKeyField, Action
import frankfurt.sql


class Meta:
    abstract : bool = False
    name: str
    table_name : str
    fields : Dict[str, BaseField] = {}
    db: Any

    def __init__(self, name : str):
        self.name = name
        self.table_name = name.lower()
        self.fields = {}

    @property
    def depends(self):
        # Return a list of methods that should be created first.

        deps = []

        for field in self.fields.values():
            if isinstance(field, ForeignKeyField):
                deps.append(field.to_field.model)

        return deps

    def extend_from_model(self, Meta):

        # Set the db.
        if hasattr(Meta, 'db'):
            self.db = Meta.db

        # Change the abstract.
        self.abstract = getattr(Meta, "abstract", self.abstract)

        # Change the table name.
        if hasattr(Meta, 'table_name'):
            self.table_name = Meta.table_name

    def __repr__(self):
        return "<MetaInfo for='{}' fields={} />".format(
            self.name,
            str(self.fields)
        )


class BaseModel:
    def __init__(self, **kwargs):

        # Save the value of the fields in this dict
        self.__data__ = {}

        # Check for the kwargs in the fields.
        for name, value in kwargs.items():
            if name in self._meta.fields:
                self.__data__[name] = value
            else:
                raise HasNotFieldTypeError(name, self)

        # Continue with default values.
        for field_name, field in self._meta.fields.items():
            if field_name not in self.__data__:
                if field.has_default:
                    self.__data__[field_name] = field.default

        # Check for not nulls (or maybe types?)
        for field_name, field in self._meta.fields.items():
            if field_name in self.__data__:
                if field.not_null and self.__data__[field_name] is None:
                    raise TypeError(
                        "Field {} cannot be null (None)".format(field_name)
                    )

    def __setitem__(self, key : str, value):
        if key not in self._meta.fields:
            msg = "Model {} has not field {}.".format(self._meta.name, key)
            raise KeyError(msg)
        self.__data__[key] = value

    def __getitem__(self, key : str):
        if key not in self._meta.fields:
            msg = "Model {} has not field {}.".format(self._meta.name, key)
            raise KeyError(msg)

        if key not in self.__data__:
            raise KeyError("Field {} has no value.".format(key))

        return self.__data__[key]

    async def save(self, conn=None):
        ins = frankfurt.sql.Insert(
            table_name=self._meta.table_name,
            values=self.__data__,
            returning=self._meta.fields.keys()
        )

        # Insert and get the record.
        record = await ins.execute(conn=conn)

        for key, value in record.items():
            self.__data__[key] = value


class ModelType(type):

    def __new__(cls, name: str, bases: Tuple[Type, ...], attrs: Dict[str, Any]):

        # Start with a fresh meta.
        meta = Meta(name)

        # Extract some information from bases.
        for base in bases:
            if hasattr(base, '_meta'):

                # Copy the reference to a database.
                meta.db = base._meta.db

                # Copy fields.
                for field_name, field in base._meta.fields.items():
                    meta.fields[field_name] = field.copy()

        if "Meta" in attrs:
            meta.extend_from_model(attrs["Meta"])

        # Append the fields defined in the model.
        for name, value in attrs.items():
            if isinstance(value, BaseField):
                meta.fields[name] = value

        # Append the meta information to the class.
        attrs["_meta"] = meta

        model = super().__new__(cls, name, bases, attrs)

        # Add the model to the dictionary of models if is not abstract.
        if not meta.abstract:
            meta.db.models[meta.table_name] = model

        # Add a reference for the fields.
        for field in meta.fields.values():
            field.model = model

        return model


class Database:
    models : Dict[str, Type[BaseModel]]
    model_class : Type[BaseModel]

    def __init__(self):

        self.models = {}

        # Declare the base model class.
        self.model_class = ModelType('BaseModel', (BaseModel, ), {
            'Meta': type('Meta', (), {
                'abstract': True,
                'db': self
            })
        })

    @property
    def Model(self):
        return self.model_class

    async def create_table(self, table_name, conn=None):

        if isinstance(table_name, ModelType):
            table_name = table_name._meta.table_name

        # Get the model.
        model = self.models[table_name]

        # Columns
        columns = []

        for field_name, field in model._meta.fields.items():

            # Resolve the constraint.
            constraint = frankfurt.sql.ColumnConstraint(
                not_null=field.not_null,
                unique=field.unique,
                primary_key=field.primary_key
            )

            # Resolve any dependecies.
            if isinstance(field, ForeignKeyField):
                reftable, refcolumn = field.refs()

                # Get the on_delete action.
                on_delete = None
                if isinstance(field.on_delete, Action):
                    on_delete = field.on_delete.value

                references = frankfurt.sql.ColumnConstraintReference(
                    reftable, refcolumn=refcolumn,
                    on_delete=on_delete
                )

                constraint.references = references

            # Append a new column.
            columns.append(frankfurt.sql.Column(
                field_name, field.__postgresql_type__(), constraint
            ))

        # CreateTable.
        stmt = frankfurt.sql.CreateTable(table_name, columns)
        return await stmt.execute(conn)

    async def create_all_tables(self, conn=None):
        # Create all the tables.

        # Sort the models in the right order.
        models_unsorted = list(self.models.keys())
        models_sorted = []

        while len(models_unsorted) > 0:
            key = models_unsorted[0]
            deps = self.models[key]._meta.depends
            if all(_._meta.table_name in models_sorted for _ in deps):
                models_sorted.append(models_unsorted.pop(0))
            else:
                models_unsorted.append(models_unsorted.pop(0))

        # Run all the statements inside a transaction.
        async with conn.transaction():
            for table_name in models_sorted:
                await self.create_table(table_name, conn=conn)


class HasNotFieldTypeError(TypeError):
    def __init__(self, field_name, cls):
        super().__init__("Model {} has not field {}".format(
            cls.__class__.__name__, field_name
        ))


db = Database()

Model = db.Model



import pytest
from mock import AsyncMock


@pytest.mark.asyncio
async def test_create_table_no_columns():
    import frankfurt.sql

    conn = AsyncMock()

    ct = frankfurt.sql.CreateTable('test', {})
    await ct.execute(conn=conn)

    conn.execute.assert_called_once_with(
        "CREATE TABLE IF NOT EXISTS test"
    )


@pytest.mark.asyncio
async def test_create_table_with_many_columns():
    import frankfurt.sql

    conn = AsyncMock()

    columns = [
        frankfurt.sql.Column('col_0', f'dt_0'),
        frankfurt.sql.Column('col_1', f'dt_1')
    ]

    ct = frankfurt.sql.CreateTable('test', columns)
    await ct.execute(conn)
    conn.execute.assert_called_once_with(
        "CREATE TABLE IF NOT EXISTS test (col_0 dt_0, col_1 dt_1)"
    )


@pytest.mark.asyncio
async def test_simple_insert():
    import frankfurt.sql

    conn = AsyncMock()
    conn.fetchrow.return_value = {'a' : 1, 'b': 2}

    ins = frankfurt.sql.Insert('test', values={'a': 1}, returning=['a', 'b'])
    record = await ins.execute(conn=conn)

    assert record['a'] == 1
    assert record['b'] == 2

    conn.fetchrow.assert_called_with(
        "INSERT INTO test (a) VALUES ($1) RETURNING (a, b)", 1
    )



import logging
from typing import Dict, Optional, List, Any
from dataclasses import dataclass


logger = logging.getLogger(__name__)


class Statement:
    pass


@dataclass
class ColumnConstraintReference:
    reftable: str
    refcolumn: Optional[str] = None
    on_delete: Optional[str] = None
    on_update: Optional[str] = None

    def __str__(self):
        stmt = f"REFERENCES {self.reftable}"

        if self.refcolumn:
            stmt = f"{stmt} ({self.refcolumn})"

        if self.on_delete:
            stmt = f"{stmt} ON DELETE {self.on_delete}"

        return stmt


@dataclass
class ColumnConstraint:
    not_null: bool = False
    primary_key: bool = False
    references: Optional[ColumnConstraintReference] = None
    unique : bool = False

    def __bool__(self):

        if self.not_null:
            return True

        if self.primary_key:
            return True

        if self.references:
            return True

        if self.unique:
            return True

        return False

    def __str__(self):

        constraints = []

        if self.not_null:
            constraints.append("NOT NULL")

        if self.unique:
            constraints.append("UNIQUE")

        if self.primary_key:
            constraints.append("PRIMARY KEY")

        if self.references:
            constraints.append(str(self.references))

        return " ".join(constraints)


@dataclass
class Column:

    name : str
    data_type : str
    constraint : Optional[ColumnConstraint] = None

    def __init__(
            self,
            name : str,
            data_type : str,
            constraint : Optional[ColumnConstraint] = None
    ):

        self.name = name
        self.data_type = data_type
        self.constraint = constraint

    def __str__(self):
        stmt = f"{self.name} {self.data_type}"
        if self.constraint:
            stmt = f"{stmt} {self.constraint}"
        return stmt


class WhereCondition:
    def __bool__(self):
        return False


class Expression:
    def __bool__(self):
        return False


class Select(Statement):
    """ Select statement. """

    select: Expression
    where: WhereCondition
    from_items: List[Any]

    async def execute(self, conn=None):
        stmt = "SELECT"

        if self.selectable:
            stmt = f"SELECT {self.select}"

        if self.from_items:
            from_items = ", ".join(str(item) for item in self.from_item)
            stmt = f"{stmt} FROM {from_items}"

        if self.where:
            stmt = f"{stmt} WHERE {self.where}"

        logger.debug(stmt)
        return await conn.fetch(stmt)


class Insert(Statement):
    """ Insert statement. """

    table_name : str
    values : Dict[str, Any]
    returning: List[str]

    def __init__(
            self, table_name : str,
            values : Optional[Dict[str, Any]] = None,
            returning : Optional[List[str]] = None
    ):

        self.table_name = table_name
        self.values = values if values is not None else {}
        self.returning = returning if returning is not None else []

    async def execute(self, conn):

        stmt = f"INSERT INTO {self.table_name}"

        values = []

        if self.values:
            cols = []
            placeholders = []

            for i, (name, value) in enumerate(self.values.items()):
                cols.append(name)
                placeholders.append(f'${i + 1}')
                values.append(value)

            cols = ', '.join(cols)
            placeholders = ', '.join(placeholders)
            stmt = f"{stmt} ({cols}) VALUES ({placeholders})"

        if self.returning:
            returning = ', '.join(self.returning)
            stmt = f"{stmt} RETURNING ({returning})"

        logger.debug(stmt)
        return await conn.fetchrow(stmt, *values)


class CreateTable(Statement):

    name : str
    columns : List[Column]

    def __init__(self, name : str, columns : List[Column]):
        self.name = name
        self.columns = columns

    async def execute(self, conn):
        stmt = f"CREATE TABLE IF NOT EXISTS {self.name}"

        if self.columns:
            columns = ', '.join(str(col) for col in self.columns)
            stmt = f"{stmt} ({columns})"

        logger.debug(stmt)
        return await conn.execute(stmt)



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="frankfurt",
    version="0.1b2",
    author="Jorge E. Cardona",
    author_email="jorgeecardona@gmail.com",
    description="Frankfurt is an ORM based on asyncpg.",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/jorgeecardona/frankfurt",
    packages=setuptools.find_packages(),
    install_requires=[
        'asyncpg',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Framework :: AsyncIO",
        "Topic :: Database",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)"
    ],
    python_requires='>=3.7',
)



from enum import Enum


class Action(Enum):
    NO_ACTION = "NO ACTION"
    RESTRICT = "RESTRICT"
    CASCADE = "CASCADE"
    SET_NULL = "SET NULL"
    SET_DEFAULT = "SET DEFAULT"


# Export the actions.
CASCADE = Action.CASCADE


class BaseField:

    not_null : bool = False
    unique : bool = False
    primary_key : bool = False

    __kwargs__ = [
        'not_null',
        'unique',
        'primary_key',
        'default'
    ]

    @property
    def default(self):
        if callable(self._default):
            return self._default()
        return self._default

    def __init__(self, **kwargs):

        for v in kwargs:
            if v not in self.__kwargs__:
                raise TypeError(f"Argument {v} is not needed.")

        self._copy_kwargs = kwargs.copy()

        # Can this field be null.
        if 'not_null' in kwargs:
            self.not_null = kwargs['not_null']

        # Is this field unique.
        if 'unique' in kwargs:
            self.unique = kwargs['unique']

        # Is this field a primary key?
        if 'primary_key' in kwargs:
            self.primary_key = kwargs['primary_key']

        # Add None as default if necessary.
        if not self.not_null and 'default' not in kwargs:
            self.has_default = True
            self._default = None
        elif 'default' in kwargs:
            self.has_default = True
            self._default = kwargs['default']
        else:
            self.has_default = False

    def copy(self):
        return self.__class__(**self._copy_kwargs)


class CharField(BaseField):
    """String-based field."""

    def __postgresql_type__(self):
        return f"VARCHAR({self.max_length})"

    def copy(self):
        return CharField(max_length=self.max_length, **self._copy_kwargs)

    def __init__(self, max_length: int = 256, **kwargs):
        super().__init__(**kwargs)
        self.max_length = max_length


class UUIDField(BaseField):
    def __postgresql_type__(self):
        return "UUID"


class BinaryField(BaseField):
    def __postgresql_type__(self):
        return "BYTEA"


class DateTimeField(BaseField):
    def __postgresql_type__(self):
        return "TIMESTAMPTZ"

    def __init__(self, auto_now=False, auto_now_add=False, **kwargs):
        super().__init__(**kwargs)

        # auto_now implies auto_now_add
        self.auto_now_add = auto_now_add or auto_now
        self.auto_now = auto_now

    def copy(self):
        return DateTimeField(
            auto_now=self.auto_now,
            auto_now_add=self.auto_now_add,
            **self._copy_kwargs
        )


class ForeignKeyField(BaseField):

    _to_field = None

    def __postgresql_type__(self):
        return self.to_field.__postgresql_type__()

    @property
    def to_field(self):
        self._resolve()
        return self._to_field

    def refs(self):
        reftable = self.to_field.model._meta.table_name
        refcolumn = None
        for field_name, field in self.to_field.model._meta.fields.items():
            if field == self.to_field:
                refcolumn = field_name

        return reftable, refcolumn

    def _resolve(self):

        if self._to_field is not None:
            return

        to_parts = self.to.split('.')
        table_name = to_parts[0]
        to_model = None

        # Resolve first the model.
        if table_name not in self.model._meta.db.models:
            raise Exception("ForeignKey can't be resolved.")
        to_model = self.model._meta.db.models[table_name]

        # Check if there is a field.
        if len(to_parts) > 1:
            to_field_name = to_parts[1]
            if to_field_name not in to_model._meta.fields:
                raise Exception("ForeignKey can't be resolved.")
            self.to_field = to_model._meta[to_field_name]
        else:
            for field_name, field in to_model._meta.fields.items():
                if field.primary_key:
                    self._to_field = field
                    break

        if self._to_field is None:
            raise Exception("ForeignKey can't be resolved.")

    def __init__(self, to, on_delete=None, **kwargs):
        super().__init__(**kwargs)
        self.to = to
        self.on_delete = on_delete

        if isinstance(self.to, BaseField):
            self._to_field = self.to

    def copy(self):
        return self.__class__(
            to=self.to,
            on_delete=self.on_delete,
            **self._copy_kwargs
        )



import pytest
from mock import AsyncMock, MagicMock, call


def test_simple_model():

    from frankfurt.models import Model
    from frankfurt import fields

    class Example(Model):
        text = fields.CharField(max_length=200)

    example = Example(text='test')
    assert example['text'] == 'test'

    with pytest.raises(KeyError):
        example['text_3']


def test_exception_incorrect_argument():
    from frankfurt.models import Model, HasNotFieldTypeError
    from frankfurt import fields

    class Example(Model):
        text = fields.CharField(max_length=200)

    with pytest.raises(HasNotFieldTypeError):
        Example(anoter='test')


def test_abstract_model():

    from frankfurt.models import Model

    class Example(Model):

        class Meta:
            abstract = True

    assert Example._meta.abstract


def test_inherit_fields():
    from frankfurt import Database
    from frankfurt import fields

    db = Database()

    class Writer(db.Model):
        pk = fields.UUIDField(primary_key=True)
        full_name = fields.CharField(max_length=400)

        class Meta:
            table_name = 'writers'

    class BaseModel(db.Model):
        pk = fields.UUIDField(primary_key=True)
        creation_date = fields.DateTimeField(auto_now_add=True)
        writer = fields.ForeignKeyField(to='writers')

        class Meta:
            abstract = True

    class Book(BaseModel):
        title = fields.CharField(max_length=200)


    # Check that book has the right fields.
    assert 'pk' in Book._meta.fields
    assert 'creation_date' in Book._meta.fields
    assert 'writer' in Book._meta.fields

    # Check that the fields are new instances.
    assert BaseModel._meta.fields['pk'] != Book._meta.fields['pk']
    assert BaseModel._meta.fields['pk'].model is BaseModel
    assert Book._meta.fields['pk'].model is Book


@pytest.mark.asyncio
async def test_create_table_from_model():
    from frankfurt.models import Model
    from frankfurt import fields

    conn = AsyncMock()

    class Writer(Model):
        pk = fields.UUIDField(primary_key=True)
        full_name = fields.CharField(max_length=400)

        class Meta:
            table_name = 'writers'

    class Book(Model):
        title = fields.CharField(max_length=200)
        date = fields.DateTimeField()
        author = fields.ForeignKeyField(
            to='writers', on_delete=fields.CASCADE
        )

        class Meta:
            table_name = 'books'

    await Book._meta.db.create_table(Book, conn=conn)

    # Assert the connection.
    conn.execute.assert_called_once_with(
        f'CREATE TABLE IF NOT EXISTS books (title VARCHAR(200), '
        'date TIMESTAMPTZ, author UUID REFERENCES writers (pk) ON '
        'DELETE CASCADE)'
    )


@pytest.mark.asyncio
async def test_create_all_tables_1():

    conn = AsyncMock()
    conn.transaction = MagicMock()

    from frankfurt.models import Database
    from frankfurt import fields

    db = Database()

    class Writer(db.Model):
        pk = fields.UUIDField(primary_key=True)
        full_name = fields.CharField(max_length=400)

        class Meta:
            table_name = 'writers'

    class Book(db.Model):
        title = fields.CharField(max_length=200)
        date = fields.DateTimeField()
        author = fields.ForeignKeyField(to='writers')

        class Meta:
            table_name = 'books'

    await db.create_all_tables(conn=conn)

    # Assert the connection.
    assert conn.execute.mock_calls == [
        call('CREATE TABLE IF NOT EXISTS writers '
             '(pk UUID PRIMARY KEY, full_name VARCHAR(400))'),
        call(f'CREATE TABLE IF NOT EXISTS books (title VARCHAR(200), '
             'date TIMESTAMPTZ, author UUID REFERENCES writers (pk))'),
    ]


@pytest.mark.asyncio
async def test_models_with_base_model():
    from frankfurt import Database
    from frankfurt import fields

    conn = AsyncMock()
    conn.transaction = MagicMock()

    db = Database()

    class BaseModel(db.Model):
        pk = fields.UUIDField(primary_key=True)

        class Meta:
            abstract = True

    class Writer(BaseModel):
        class Meta:
            table_name = 'writers'

    class Book(BaseModel):
        author = fields.ForeignKeyField(to='writers')

    await db.create_all_tables(conn=conn)

    assert conn.execute.mock_calls == [
        call('CREATE TABLE IF NOT EXISTS writers '
             '(pk UUID PRIMARY KEY)'),
        call(f'CREATE TABLE IF NOT EXISTS book '
             '(pk UUID PRIMARY KEY, author UUID REFERENCES writers (pk))'),
    ]
